#!/bin/sh
#
# New distribution version..
vers=`TZ=GMT date "+%y%m%d%H%M"`
cat  > version.c << EOF
/* versio.c -- autogenerated by version.sh at distribution creation */
char *version_time = "$vers";
EOF
